import kestrel.config as cfg
import os


def test_env_vars_in_config():

    test_config = """---
credentials:
  username: $TEST_USER
  password: $TEST_PASSWORD
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")

    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = cfg.load_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"


def test_env_vars_in_config_overwrite():

    test_config = """---
credentials:
  username: ${TEST_USER}
  password: ${TEST_PASSWORD}
debug:
  cache_directory_prefix: $KESTREL_CACHE_DIRECTORY_PREFIX
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")
    os.environ["KESTREL_CACHE_DIRECTORY_PREFIX"] = "Kestrel2.0-"
    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = cfg.load_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"
    assert config["debug"]["cache_directory_prefix"] == "Kestrel2.0-"

def test_empty_env_var_in_config():
    test_config = """---
credentials:
  username: ${TEST_USER}
  password: ${TEST_PASSWORD}
debug:
  cache_directory_prefix: $I_DONT_EXIST
    """
    os.environ["TEST_USER"] = "test-user"
    os.environ["TEST_PASSWORD"] = "test-password"
    os.environ["KESTREL_CONFIG"] = os.path.join(os.sep, "tmp", "config.yaml")
    os.environ["KESTREL_CACHE_DIRECTORY_PREFIX"] = "Kestrel2.0-"
    with open(os.getenv("KESTREL_CONFIG"), "w") as fp:
        fp.write(test_config)
    config = cfg.load_config()
    assert config["credentials"]["username"] == "test-user"
    assert config["credentials"]["password"] == "test-password"
    assert config["debug"]["cache_directory_prefix"] == "$I_DONT_EXIST"



import json
import os
import pytest

from kestrel.session import Session

from .utils import set_empty_kestrel_config, set_no_prefetch_kestrel_config


@pytest.fixture
def fake_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "../../../test-data/test_bundle.json")


@pytest.fixture
def proc_bundle_file():
    cwd = os.path.dirname(os.path.abspath(__file__))
    return os.path.join(cwd, "../../../test-data/doctored-1k.json")


def test_return_table_not_exist(set_empty_kestrel_config, fake_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        from file://{fake_bundle_file}
                        where [network-traffic:dst_port = 22]
                procs = FIND process CREATED conns
                """
        summaries = s.execute(stmt)
    correct_dict = {
        "display": "execution summary",
        "data": {
            "variables updated": [
                {
                    "VARIABLE": "conns",
                    "TYPE": "network-traffic",
                    "#(ENTITIES)": 29,
                    "#(RECORDS)": 29,
                    "ipv4-addr*": 58,
                    "network-traffic*": 0,
                    "user-account*": 29,
                },
                {
                    "VARIABLE": "procs",
                    "TYPE": "process",
                    "#(ENTITIES)": 0,
                    "#(RECORDS)": 0,
                    "ipv4-addr*": 0,
                    "network-traffic*": 0,
                    "user-account*": 0,
                },
            ],
            "footnotes": ["*Number of related records cached."],
        },
    }
    output_dict = summaries[0].to_dict()
    del output_dict["data"]["execution time"]
    assert output_dict == correct_dict


def test_find_srcs(set_empty_kestrel_config, fake_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        from file://{fake_bundle_file}
                        where dst_port = 22
                srcs = FIND ipv4-addr CREATED conns
                """
        s.execute(stmt)
        srcs = s.get_variable("srcs")
        assert len(srcs) == 24


def test_find_srcs_limit(set_empty_kestrel_config, fake_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        from file://{fake_bundle_file}
                        where dst_port = 22
                srcs = FIND ipv4-addr CREATED conns LIMIT 1
                """
        s.execute(stmt)
        srcs = s.get_variable("srcs")
        assert len(srcs) == 1


def test_find_ips_linked_to_process_limit_1(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where name = 'svchost.exe'
                ips = FIND ipv4-addr LINKED procs LIMIT 1
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 704
        ips = s.get_variable("ips")
        print(json.dumps(ips, indent=4))
        # FIXME: the LIMIT does not work precisely (OK to leave it for now)
        assert len(ips) == 2


def test_find_ips_linked_to_process_limit_10(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where name = 'svchost.exe'
                ips = FIND ipv4-addr LINKED procs LIMIT 10
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 704
        ips = s.get_variable("ips")
        print(json.dumps(ips, indent=4))
        assert len(ips) == 6


def test_find_file_linked_to_process(set_empty_kestrel_config):
    stixshifter_data_url = "https://raw.githubusercontent.com/opencybersecurityalliance/stix-shifter/develop/data/cybox"
    bundle = f"{stixshifter_data_url}/carbon_black/cb_observed_156.json"
    with Session() as s:
        stmt = f"""
                procs = get process
                        from {bundle}
                        where name = 'svctest.exe'
                files = FIND file LINKED procs
                """
        s.execute(stmt)
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 3


def test_find_ips_linked_to_process_limit(set_empty_kestrel_config):
    stixshifter_data_url = "https://raw.githubusercontent.com/opencybersecurityalliance/stix-shifter/develop/data/cybox"
    bundle = f"{stixshifter_data_url}/carbon_black/cb_observed_156.json"
    with Session() as s:
        stmt = f"""
                procs = get process
                        from {bundle}
                        where name = 'svctest.exe'
                ips = FIND ipv4-addr LINKED procs LIMIT 10
                """
        s.execute(stmt)
        ips = s.get_variable("ips")
        print(json.dumps(ips, indent=4))
        assert len(ips) == 6


def test_find_file_loaded_by_process_wmic(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                files = FIND file LOADED BY procs
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 7
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 1


def test_find_file_linked_to_process_wmic(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                files = FIND file LINKED procs
                """
        s.execute(stmt)
        procs = s.get_variable("procs")
        print(json.dumps(procs, indent=4))
        assert len(procs) == 7
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 2


def test_find_process_created_process_wmic(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where command_line LIKE 'wmic%'
                parents = FIND process CREATED procs
                """
        s.execute(stmt)
        data = s.get_variable("parents")
        print(json.dumps(data, indent=4))
        assert len(data)


def test_find_refs_resolution_not_reversed_src_ref(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                nt = get network-traffic
                     from file://{proc_bundle_file}
                     where src_port > 0
                p = FIND process CREATED nt
                """
        s.execute(stmt)
        p = s.get_variable("p")
        # assert len(p) == 948  # grep -c opened_connection_refs tests/doctored-1k.json
        assert len(p) >= 948  # FIXME: duplicate process objects


def test_find_refs_resolution_not_reversed_src_ref_limit(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                nt = get network-traffic
                     from file://{proc_bundle_file}
                     where src_port > 0
                p = FIND process CREATED nt LIMIT 10
                """
        s.execute(stmt)
        p = s.get_variable("p")
        # assert len(p) == 948  # grep -c opened_connection_refs tests/doctored-1k.json
        assert len(p) == 10
 

def test_find_refs_resolution_reversed_src_ref(set_empty_kestrel_config, proc_bundle_file):
    with Session(debug_mode=True) as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where name LIKE '%'
                conns = FIND network-traffic CREATED BY procs
                """
        s.execute(stmt)
        conns = s.get_variable("conns")
        assert (
            len(conns) == 853
        )  # FIXME: should be 948, I think (id collisions for network-traffic)

        # DISP with a ref (parent_ref) and ambiguous column (command_line)
        disp_out = s.execute("DISP procs ATTR name, parent_ref.name, command_line")
        data = disp_out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))


def test_find_refs_resolution_reversed_src_ref_limit(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                procs = get process
                        from file://{proc_bundle_file}
                        where name LIKE '%'
                conns = FIND network-traffic CREATED BY procs LIMIT 10
                """
        s.execute(stmt)
        conns = s.get_variable("conns")
        assert (
            len(conns) == 853
        )

        # DISP with a ref (parent_ref) and ambiguous column (command_line)
        disp_out = s.execute("DISP procs ATTR name, parent_ref.name, command_line")
        data = disp_out[0].to_dict()["data"]
        print(json.dumps(data, indent=4))


def test_find_without_where_ext_pattern(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                procs = FIND process CREATED conns
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 193
        assert conns.records_count == 203

        assert len(procs) == 471
        assert procs.records_count == 471


# stix_bundle connector does not support extended graph
# disable prefetch to test
def test_find_with_where_ext_pattern(set_no_prefetch_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                procs = FIND process CREATED conns
                        WHERE ipv4-addr:value = '192.168.1.1'
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 193
        assert conns.records_count == 203

        assert len(procs) == 203
        assert procs.records_count == 203


def test_find_with_limit(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                procs = FIND process CREATED conns LIMIT 100
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 193
        assert conns.records_count == 203

        assert len(procs) == 100
        assert procs.records_count == 100


def test_find_with_where_centered_pattern(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '127.0.0.1'

                procs = FIND process CREATED conns
                        WHERE name = 'vmware.exe'
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 193
        assert conns.records_count == 203

        assert len(procs) == 1
        assert procs.records_count == 1


def test_find_from_empty_input(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                conns = get network-traffic
                        FROM file://{proc_bundle_file}
                        WHERE network-traffic:src_ref.value = '100.0.0.1'

                procs = FIND process CREATED conns
                        WHERE name = 'vmware.exe'
                """
        s.execute(stmt)

        conns = s.symtable["conns"]
        procs = s.symtable["procs"]

        assert len(conns) == 0
        assert conns.records_count == 0

        assert len(procs) == 0
        assert procs.records_count == 0


def test_find_dir_contained_file(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                files = GET file
                        FROM file://{proc_bundle_file}
                        WHERE name = 'svchost.exe'
                dirs = FIND directory CONTAINED files
                """
        s.execute(stmt)
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        dirs = s.get_variable("dirs")
        print(json.dumps(dirs, indent=4))
        assert len(dirs) == 1


def test_find_files_contained_by_dir(set_empty_kestrel_config, proc_bundle_file):
    with Session() as s:
        stmt = f"""
                dirs = GET directory
                       FROM file://{proc_bundle_file}
                       WHERE path = 'C:\\Windows\\System32'
                files = FIND file CONTAINED BY dirs
                """
        s.execute(stmt)
        files = s.get_variable("files")
        print(json.dumps(files, indent=4))
        assert len(files) == 10



import pathlib
import os
import uuid
import collections.abc
from typeguard import typechecked
from typing import Union, Iterable, Mapping
import logging


@typechecked
def unescape_quoted_string(s: str):
    if s.startswith("r"):
        return s[2:-1]
    else:
        return s[1:-1].encode("utf-8").decode("unicode_escape")


@typechecked
def lowered_str_list(xs: Iterable):
    return [x.lower() for x in xs if isinstance(x, str)]


@typechecked
def mask_value_in_nested_dict(d: Mapping, sensitive_branch: str):
    # sensitive_branch is the key of the branch to be masked out
    # if sensitive_branch == '*', then mask all values in the branch
    # if not, locate the sensitive branch and masks all values in that branch
    if d:
        for k, v in d.items():
            if k == sensitive_branch:
                sensitive_branch = "*"
            if isinstance(v, collections.abc.Mapping):
                d[k] = mask_value_in_nested_dict(v, sensitive_branch)
            elif isinstance(v, str) and sensitive_branch == "*":
                d[k] = "********"
    return d


@typechecked
def update_nested_dict(dict_old: Mapping, dict_new: Union[Mapping, None]):
    if dict_new:
        for k, v in dict_new.items():
            if isinstance(v, collections.abc.Mapping) and k in dict_old:
                dict_old[k] = update_nested_dict(dict_old[k], v)
            else:
                dict_old[k] = v
    return dict_old


@typechecked
def remove_empty_dicts(ds: Iterable[Mapping]):
    # remove dict with all values as None in list({string:string})
    # this is the results from SQL query
    return [d for d in ds if set(d.values()) != {None}]


@typechecked
def dedup_dicts(ds: Iterable[Mapping]):
    # deduplicate list({string:string})
    # this is the results from SQL query
    return [dict(s) for s in set(frozenset(d.items()) for d in ds)]


@typechecked
def dedup_ordered_dicts(ds: Iterable[Mapping]):
    # deduplicate list({string:string})
    # maintain the order if seen
    res = []
    seen = set()
    for d in ds:
        s = str(d)
        if s not in seen:
            res.append(d)
        seen.add(s)
    return res


def subgroup_list(xs, gsize):
    return [xs[i : i + gsize] for i in range(0, len(xs), gsize)]


def mkdtemp():
    # create a temporary directory and return it (named after a random uuid)
    ps = None
    while not ps or pathlib.Path(ps).exists():
        ps = str(uuid.uuid4())
    p = pathlib.Path(ps)
    p.mkdir(parents=True, exist_ok=True)
    return p


def add_logging_handler(handler, if_debug):
    fmt = "%(asctime)s %(levelname)s %(name)s %(message)s"
    datefmt = "%H:%M:%S"
    formatter = logging.Formatter(fmt=fmt, datefmt=datefmt)

    handler.setFormatter(formatter)

    root_logger = logging.getLogger()
    current_logging_level = root_logger.getEffectiveLevel()
    root_logger.addHandler(handler)
    root_logger.setLevel(logging.DEBUG if if_debug else logging.INFO)

    return handler, current_logging_level


def clear_logging_handlers():
    root_logger = logging.getLogger()
    for h in root_logger.handlers[:]:
        root_logger.removeHandler(h)
        h.close()


def resolve_path_in_kestrel_env_var():
    for key in os.environ:
        if key.startswith("KESTREL") or key.startswith("kestrel"):
            path = os.environ[key]
            if os.path.exists(path):
                os.environ[key] = resolve_path(path)


def resolve_path(path):
    return os.path.abspath(os.path.expanduser(os.path.expandvars(path)))


class set_current_working_directory:
    def __init__(self, new_cwd):
        self.tmp_cwd = new_cwd

    def __enter__(self):
        self.cwd = os.getcwd()
        os.chdir(self.tmp_cwd)

    def __exit__(self, exception_type, exception_value, traceback):
        os.chdir(self.cwd)



from pkgutil import get_data
from importlib import resources


def load_data_file(package_name, file_name):
    try:
        # resources.files() is introduced in Python 3.9
        content = resources.files(package_name).joinpath(file_name).read_text()
    except AttributeError:
        # Python 3.8; deprecation warning forward
        content = get_data(package_name, file_name).decode("utf-8")

    return content



import os
import pytest

import kestrel
from kestrel.session import Session
from kestrel.codegen import prefetch

from .utils import set_empty_kestrel_config, set_no_prefetch_kestrel_config


def test_default_prefetch_config(set_empty_kestrel_config):
    with Session() as s:
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "get", "ipv4-addr"
            )
            == True
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "network-traffic"
            )
            == True
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "process"
            )
            == True
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "file"
            )
            == False
        )


def test_prefetch_disabled_config(set_no_prefetch_kestrel_config):
    with Session() as s:
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "get", "ipv4-addr"
            )
            == False
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "network-traffic"
            )
            == False
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "process"
            )
            == False
        )
        assert (
            prefetch._is_prefetch_allowed_in_config(
                s.config["prefetch"], "find", "file"
            )
            == False
        )


def test_env_var_resolve(tmp_path):
    os.chdir(tmp_path)
    config_name = "abc.yaml"
    with open(config_name, "w") as config:
        config.write(r"""
language:
  default_variable: "_"
""")
    os.environ[kestrel.config.CONFIG_PATH_ENV_VAR] = config_name
    s = Session()
    full_path = os.path.join(os.getcwd(), config_name)
    assert os.environ[kestrel.config.CONFIG_PATH_ENV_VAR] == full_path 
